@@ -2,16 +2,16 @@
 Provides classes and functions for the end-to-end encryption mode.
 """
 import abc
-import asyncio
 import binascii
+import collections
 import datetime
 import enum
+import functools
 import hashlib
 import hmac
 import json
 import mimetypes
 import os
-import ssl
 import struct
 from typing import Tuple
 
@@ -20,7 +20,6 @@ import libnacl.encode
 import libnacl.public
 import libnacl.secret
 from aiohttp import web
-from aiohttp.web_urldispatcher import UrlDispatcher
 
 from . import ReceptionCapability
 from .exception import (
@@ -35,13 +34,17 @@ from .util import (
     AioRunMixin,
     ViewIOReader,
     ViewIOWriter,
-    aio_run_proxy_decorator,
+    aio_run_proxy,
     randint,
 )
 
 __all__ = (
     'BLOB_ID_LENGTH',
-    'AbstractCallback',
+    'MAX_HTTP_REQUEST_SIZE',
+    'CallbackContext',
+    'handle_callback',
+    'create_application',
+    'add_callback_route',
     'Message',
     'DeliveryReceipt',
     'TextMessage',
@@ -52,6 +55,11 @@ __all__ = (
 
 BLOB_ID_LENGTH = 16
 
+# A box can contain up to 4000 bytes, so this should be sufficient.
+# The remaining POST parameters aren't that big.
+# See: https://gateway.threema.ch/en/developer/api
+MAX_HTTP_REQUEST_SIZE = 8192
+
 
 def _pk_encrypt(key_pair: Tuple[Key, Key], data: bytes, nonce: bytes = None):
     """
@@ -96,7 +104,6 @@ def _pk_decrypt(key_pair: Tuple[Key, Key], nonce: bytes, data: bytes):
     return box.decrypt(data, nonce=nonce)
 
 
-# TODO: Raises?
 def _sk_encrypt(key, data, nonce=None):
     """
     Encrypt data by using secret-key encryption.
@@ -119,7 +126,6 @@ def _sk_encrypt(key, data, nonce=None):
     return data[:nonce_length], data[nonce_length:]
 
 
-# TODO: Raises?
 def _sk_decrypt(key, nonce, data):
     """
     Decrypt data by using secret-key decryption.
@@ -136,82 +142,32 @@ def _sk_decrypt(key, nonce, data):
     return box.decrypt(data, nonce=nonce)
 
 
-# TODO: Add logging
-# TODO: Add docstrings
-class AbstractCallback(metaclass=abc.ABCMeta):
-    """
-    Raises :exc:`TypeError` in case no valid certificate has been
-    provided.
-    """
+def _validate_hmac(encoded_secret, expected_mac, response):
+    try:
+        message = ''.join((
+            response['from'],
+            response['to'],
+            response['messageId'],
+            response['date'],
+            response['nonce'],
+            response['box'],
+        )).encode('ascii')
+    except UnicodeError as exc:
+        raise CallbackError(400, 'Cannot concatenate HMAC message') from exc
+    hmac_ = hmac.new(encoded_secret, msg=message, digestmod=hashlib.sha256)
+    actual_mac = hmac_.hexdigest()
+    if not hmac.compare_digest(expected_mac, actual_mac):
+        raise CallbackError(400, 'MACs do not match')
 
-    def __init__(self, connection, loop=None, route='/gateway_callback'):
-        self.connection = connection
-        # Note: I'm guessing here the secret must be ASCII
-        self.encoded_secret = connection.secret.encode('ascii')
-        self.loop = asyncio.get_event_loop() if loop is None else loop
-        # Create router
-        self.router = self.create_router(route)
-        # Create application
-        self.application = self.create_application(self.router, loop)
-        self.handler = self.create_handler()
-        self.server = None
 
-    # noinspection PyMethodMayBeStatic
-    def create_ssl_context(self, certfile, keyfile=None):
-        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
-        ssl_context.load_cert_chain(certfile=certfile, keyfile=keyfile)
-        return ssl_context
+CallbackContext = collections.namedtuple('CallbackContext', [
+    'encoded_secret', 'connection', 'message_handler', 'receive_handler',
+])
 
-    def create_router(self, route):
-        self.route = route
-        router = UrlDispatcher()
-        router.add_route('POST', route, self._handle_and_catch_error)
-        return router
 
-    # noinspection PyMethodMayBeStatic
-    def create_application(self, router, loop):
-        # A box can contain up to 4000 bytes, so this should be sufficient.
-        # The remaining POST parameters aren't that big.
-        # See: https://gateway.threema.ch/en/developer/api
-        request_size_max = 8192
-        return web.Application(router=router, loop=loop, client_max_size=request_size_max)
-
-    def create_handler(self):
-        return self.application.make_handler()
-
-    @asyncio.coroutine
-    def create_server(self, certfile, keyfile=None, host=None, port=443, **kwargs):
-        # Create SSL context
-        ssl_context = self.create_ssl_context(certfile, keyfile=keyfile)
-        # Create server
-        # noinspection PyArgumentList
-        server = yield from self.loop.create_server(
-            self.handler, host=host, port=port, ssl=ssl_context, **kwargs)
-        return server
-
-    @asyncio.coroutine
-    def close(self, timeout=10.0):
-        # Stop handler and application
-        yield from self.application.shutdown()
-        yield from self.handler.shutdown(timeout=timeout)
-        yield from self.application.cleanup()
-
-    @asyncio.coroutine
-    def _handle_and_catch_error(self, request):
-        try:
-            return (yield from self.handle_callback(request))
-        except CallbackError as exc:
-            # TODO: Log
-            # Note: For security reasons, we do not send the reason
-            return web.Response(status=exc.status)
-        except Exception:
-            # TODO: Log
-            raise
-
-    # TODO: Raises?
-    @asyncio.coroutine
-    def handle_callback(self, request):
-        response = yield from request.post()
+async def handle_callback(context, request):
+    try:
+        response = await request.post()
 
         # Unpack fields
         try:
@@ -225,9 +181,10 @@ class AbstractCallback(metaclass=abc.ABCMeta):
         except KeyError as exc:
             raise CallbackError(400, 'Could not unpack required fields') from exc
 
-        # Validate MAC and ID
-        self.validate_mac(mac, response)
-        self.validate_id(to_id)
+        # Validate HMAC and ID
+        _validate_hmac(context.encoded_secret, mac, response)
+        if to_id != context.connection.id:
+            raise CallbackError(400, 'IDs do not match')
 
         # Validate from id length
         if len(from_id) != 8:
@@ -252,7 +209,7 @@ class AbstractCallback(metaclass=abc.ABCMeta):
 
         # Unpack message
         try:
-            message = yield from Message.receive(self.connection, {
+            message = await context.receive_handler(context.connection, {
                 'from_id': from_id,
                 'message_id': message_id,
                 'date': date,
@@ -261,41 +218,36 @@ class AbstractCallback(metaclass=abc.ABCMeta):
             raise CallbackError(400, str(exc)) from exc
 
         # Pass message to handler
-        try:
-            yield from self.receive_message(message)
-        except TypeError:
-            # TODO: Log error that the inherited method 'receive_message' MUST
-            #       be a coroutine.
-            raise
+        await context.message_handler(message)
 
         # Respond with 'OK'
         return web.Response(status=200)
+    except CallbackError as exc:
+        # Note: For security reasons, we do not send the reason
+        return web.Response(status=exc.status)
+    except Exception:
+        raise
 
-    def validate_mac(self, expected_mac, response):
-        try:
-            message = ''.join((
-                response['from'],
-                response['to'],
-                response['messageId'],
-                response['date'],
-                response['nonce'],
-                response['box'],
-            )).encode('ascii')
-        except UnicodeError as exc:
-            raise CallbackError(400, 'Cannot concatenate HMAC message') from exc
-        hmac_ = hmac.new(self.encoded_secret, msg=message, digestmod=hashlib.sha256)
-        actual_mac = hmac_.hexdigest()
-        if not hmac.compare_digest(expected_mac, actual_mac):
-            raise CallbackError(400, 'MACs do not match')
 
-    def validate_id(self, to_id):
-        if to_id != self.connection.id:
-            raise CallbackError(400, 'IDs do not match')
+def create_application(connection):
+    application = web.Application(client_max_size=MAX_HTTP_REQUEST_SIZE)
+    return application
 
-    @asyncio.coroutine
-    @abc.abstractmethod
-    def receive_message(self, message):
-        raise NotImplementedError
+
+def add_callback_route(
+    connection, application, message_handler,
+    path='/gateway_callback', receive_handler=None,
+):
+    if receive_handler is None:
+        receive_handler = Message.receive
+    context = CallbackContext(
+        connection.secret.encode('ascii'),
+        connection,
+        message_handler,
+        receive_handler,
+    )
+    application.router.add_routes([web.post(
+        path, functools.partial(handle_callback, context))])
 
 
 # TODO: Update docstring (arguments)
@@ -429,8 +381,7 @@ class Message(AioRunMixin, metaclass=abc.ABCMeta):
 
     # TODO: Raises?
     @property
-    @asyncio.coroutine
-    def key(self):
+    async def key(self):
         """
         Get the public key of the recipient (outgoing messages) or the
         sender (incoming messages). Will be request from the server
@@ -440,7 +391,7 @@ class Message(AioRunMixin, metaclass=abc.ABCMeta):
         decoded if required.
         """
         if self._key is None:
-            self._key = yield from self._connection.get_public_key(self.to_id)
+            self._key = await self._connection.get_public_key(self.to_id)
         return self._key
 
     @key.setter
@@ -466,8 +417,7 @@ class Message(AioRunMixin, metaclass=abc.ABCMeta):
         self._key_file = key_file
 
     # TODO: Raises?
-    @asyncio.coroutine
-    def send(self, get_data_only=False):
+    async def send(self, get_data_only=False):
         """
         Send a message.
 
@@ -485,7 +435,7 @@ class Message(AioRunMixin, metaclass=abc.ABCMeta):
             raise MessageError('Could not pack type') from exc
 
         # Get content data
-        yield from self.pack(writer)
+        await self.pack(writer)
 
         # Generate 0 < padding < 256
         padding_length = randint(1, 255)
@@ -493,21 +443,20 @@ class Message(AioRunMixin, metaclass=abc.ABCMeta):
         writer.writeexactly(bytes([padding_length] * padding_length))
 
         # Encrypt message
-        nonce, data = yield from self.encrypt(writer.getvalue())
+        nonce, data = await self.encrypt(writer.getvalue())
         if get_data_only:
             return nonce, data
 
         # Send message
-        return (yield from self._connection.send_e2e(**{
+        return await self._connection.send_e2e(**{
             'to': self.to_id,
             'nonce': binascii.hexlify(nonce).decode('ascii'),
             'box': binascii.hexlify(data).decode('ascii')
-        }))
+        })
 
     # TODO: Raises?
     @classmethod
-    @asyncio.coroutine
-    def receive(cls, connection, parameters, nonce, data):
+    async def receive(cls, connection, parameters, nonce, data):
         """
         Return a :class:`Message` instance from an encrypted message.
 
@@ -529,7 +478,7 @@ class Message(AioRunMixin, metaclass=abc.ABCMeta):
             - :exc:`MessageError` in case the message is invalid.
         """
         # Decrypt message
-        key_pair = yield from cls.get_decrypt_key_pair(connection, parameters)
+        key_pair = await cls.get_decrypt_key_pair(connection, parameters)
         data = cls.decrypt(nonce, data, key_pair)
 
         # Unpack type and padding length
@@ -550,11 +499,10 @@ class Message(AioRunMixin, metaclass=abc.ABCMeta):
         reader = ViewIOReader(data[1:-padding_length])
 
         # Unpack message
-        return (yield from class_.unpack(connection, parameters, key_pair, reader))
+        return await class_.unpack(connection, parameters, key_pair, reader)
 
-    @asyncio.coroutine
     @abc.abstractmethod
-    def pack(self, writer):
+    async def pack(self, writer):
         """
         Pack payload data.
 
@@ -568,9 +516,8 @@ class Message(AioRunMixin, metaclass=abc.ABCMeta):
         raise NotImplementedError
 
     @classmethod
-    @asyncio.coroutine
     @abc.abstractmethod
-    def unpack(cls, connection, parameters, key_pair, reader):
+    async def unpack(cls, connection, parameters, key_pair, reader):
         """
         Return a :class:`Message` instance from raw payload data.
 
@@ -593,8 +540,7 @@ class Message(AioRunMixin, metaclass=abc.ABCMeta):
         raise NotImplementedError
 
     # TODO: Raises?
-    @asyncio.coroutine
-    def check_capabilities(self, required_capabilities):
+    async def check_capabilities(self, required_capabilities):
         """
         Test for capabilities of a recipient.
 
@@ -607,27 +553,25 @@ class Message(AioRunMixin, metaclass=abc.ABCMeta):
         """
         # Check capabilities of a recipient
         capabilities_coroutine = self._connection.get_reception_capabilities(self.to_id)
-        recipient_capabilities = yield from capabilities_coroutine
+        recipient_capabilities = await capabilities_coroutine
         if not required_capabilities <= recipient_capabilities:
             missing_capabilities = required_capabilities - recipient_capabilities
             raise MissingCapabilityError(missing_capabilities)
 
     # TODO: Raises?
     @property
-    @asyncio.coroutine
-    def get_encrypt_key_pair(self):
+    async def get_encrypt_key_pair(self):
         """
         Return a tuple containing our private key and the public key of
         the recipient.
         """
         private = self._connection.key
-        public = yield from self.key
+        public = await self.key
         return private, public
 
     # TODO: Raises?
     @classmethod
-    @asyncio.coroutine
-    def get_decrypt_key_pair(cls, connection, parameters):
+    async def get_decrypt_key_pair(cls, connection, parameters):
         """
         Return a tuple containing our private key and the public key of
         the sender.
@@ -638,11 +582,10 @@ class Message(AioRunMixin, metaclass=abc.ABCMeta):
               :method:`Message.unpack`.
         """
         private = connection.key
-        public = yield from connection.get_public_key(parameters['from_id'])
+        public = await connection.get_public_key(parameters['from_id'])
         return private, public
 
-    @asyncio.coroutine
-    def encrypt(self, data, key_pair=None, nonce=None):
+    async def encrypt(self, data, key_pair=None, nonce=None):
         """
         Encrypt data.
 
@@ -659,7 +602,7 @@ class Message(AioRunMixin, metaclass=abc.ABCMeta):
         """
         # Key pair specified?
         if key_pair is None:
-            key_pair = yield from self.get_encrypt_key_pair
+            key_pair = await self.get_encrypt_key_pair
 
         # Encrypt
         try:
@@ -690,7 +633,7 @@ class Message(AioRunMixin, metaclass=abc.ABCMeta):
 
 
 # TODO: Update docstring (arguments)
-@aio_run_proxy_decorator
+@aio_run_proxy
 class DeliveryReceipt(Message):
     """
     A delivery receipt that can be received in end-to-end mode.
@@ -708,18 +651,14 @@ class DeliveryReceipt(Message):
         'unpack',
     }
 
-    # TODO: Re-enable unique decorator and remove `user_ack` with next major version
-    # @enum.unique
+    @enum.unique
     class ReceiptType(enum.IntEnum):
         """
         Describes message receipt types.
-
-        .. warning:: `user_ack` is deprecated and will be removed with
-           the next major release. Use `user_acknowledge` instead.
         """
         received = 0x01
         read = 0x02
-        user_acknowledge = user_ack = 0x03
+        user_acknowledge = 0x03
         user_decline = 0x04
 
     def __init__(
@@ -742,8 +681,7 @@ class DeliveryReceipt(Message):
         ids = (binascii.hexlify(id_).decode('ascii') for id_ in self.message_ids)
         return 'Delivery receipt({}): {}'.format(self.receipt_type.name, ', '.join(ids))
 
-    @asyncio.coroutine
-    def pack(self, writer):
+    async def pack(self, writer):
         # Pack content
         try:
             writer.writeexactly(struct.pack('<B', self.receipt_type.value))
@@ -753,8 +691,7 @@ class DeliveryReceipt(Message):
             writer.writeexactly(message_id)
 
     @classmethod
-    @asyncio.coroutine
-    def unpack(cls, connection, parameters, key_pair, reader):
+    async def unpack(cls, connection, parameters, key_pair, reader):
         # Check length
         length = len(reader)
         if length < 9 or (length - 1) % 8 != 0:
@@ -783,7 +720,7 @@ class DeliveryReceipt(Message):
 
 
 # TODO: Update docstring (arguments)
-@aio_run_proxy_decorator
+@aio_run_proxy
 class TextMessage(Message):
     """
     A text message.
@@ -822,8 +759,7 @@ class TextMessage(Message):
     def __str__(self):
         return self.text
 
-    @asyncio.coroutine
-    def pack(self, writer):
+    async def pack(self, writer):
         # Encode text
         try:
             text = self.text.encode('utf-8')
@@ -834,8 +770,7 @@ class TextMessage(Message):
         writer.writeexactly(text)
 
     @classmethod
-    @asyncio.coroutine
-    def unpack(cls, connection, parameters, key_pair, reader):
+    async def unpack(cls, connection, parameters, key_pair, reader):
         # Get text
         text = bytes(reader.readexactly(len(reader)))
 
@@ -852,7 +787,7 @@ class TextMessage(Message):
 
 
 # TODO: Update docstring (arguments)
-@aio_run_proxy_decorator
+@aio_run_proxy
 class ImageMessage(Message):
     """
     An image message.
@@ -945,14 +880,13 @@ class ImageMessage(Message):
                 raise UnsupportedMimeTypeError(mime_type)
             self._mime_type = mime_type
 
-    @asyncio.coroutine
-    def pack(self, writer):
+    async def pack(self, writer):
         # Check capabilities of recipient
-        yield from self.check_capabilities(self.required_capabilities)
+        await self.check_capabilities(self.required_capabilities)
 
         # Encrypt and upload image
-        image_nonce, image_data = yield from self.encrypt(self.image)
-        blob_id = yield from self._connection.upload(image_data)
+        image_nonce, image_data = await self.encrypt(self.image)
+        blob_id = await self._connection.upload(image_data)
         try:
             blob_id = binascii.unhexlify(blob_id)
         except binascii.Error as exc:
@@ -968,8 +902,7 @@ class ImageMessage(Message):
         writer.writeexactly(data)
 
     @classmethod
-    @asyncio.coroutine
-    def unpack(cls, connection, parameters, key_pair, reader):
+    async def unpack(cls, connection, parameters, key_pair, reader):
         # Unpack blob id, image length and image nonce
         length = struct.calcsize(cls._formatter)
         try:
@@ -980,8 +913,8 @@ class ImageMessage(Message):
 
         # Download and decrypt image
         blob_id = binascii.hexlify(blob_id).decode('ascii')
-        response = yield from connection.download(blob_id)
-        image_data = yield from response.read()
+        response = await connection.download(blob_id)
+        image_data = await response.read()
         image = cls.decrypt(image_nonce, image_data, key_pair)
 
         # Return instance
@@ -992,7 +925,7 @@ class ImageMessage(Message):
 
 
 # TODO: Update docstring (arguments)
-@aio_run_proxy_decorator
+@aio_run_proxy
 class VideoMessage(Message):
     """
     A video message including a thumbnail.
@@ -1099,17 +1032,16 @@ class VideoMessage(Message):
                 # Read content
                 self._thumbnail_content = file.read()
 
-    @asyncio.coroutine
-    def pack(self, writer):
+    async def pack(self, writer):
         # Check capabilities of recipient
-        yield from self.check_capabilities(self.required_capabilities)
+        await self.check_capabilities(self.required_capabilities)
 
         # Generate a symmetric key for the video and its thumbnail
         key, _ = Key.generate_secret_key()
 
         # Encrypt and upload video by a newly generated symmetric key
         _, video_data = _sk_encrypt(key, self.video, nonce=self.nonce['video'])
-        video_id = yield from self._connection.upload(video_data)
+        video_id = await self._connection.upload(video_data)
         try:
             video_id = binascii.unhexlify(video_id)
         except binascii.Error as exc:
@@ -1118,7 +1050,7 @@ class VideoMessage(Message):
         # Encrypt and upload thumbnail
         _, thumbnail_data = _sk_encrypt(
             key, self.thumbnail_content, nonce=self.nonce['thumbnail'])
-        thumbnail_id = yield from self._connection.upload(thumbnail_data)
+        thumbnail_id = await self._connection.upload(thumbnail_data)
         try:
             thumbnail_id = binascii.unhexlify(thumbnail_id)
         except binascii.Error as exc:
@@ -1143,8 +1075,7 @@ class VideoMessage(Message):
         writer.writeexactly(data)
 
     @classmethod
-    @asyncio.coroutine
-    def unpack(cls, connection, parameters, key_pair, reader):
+    async def unpack(cls, connection, parameters, key_pair, reader):
         # Unpack duration, blob ids, video/thumbnail length and the secret key
         length = struct.calcsize(cls._formatter)
         try:
@@ -1156,8 +1087,8 @@ class VideoMessage(Message):
 
         # Download and decrypt thumbnail
         thumbnail_id = binascii.hexlify(thumbnail_id).decode('ascii')
-        response = yield from connection.download(thumbnail_id)
-        thumbnail_data = yield from response.read()
+        response = await connection.download(thumbnail_id)
+        thumbnail_data = await response.read()
         thumbnail_content = _sk_decrypt(key, cls.nonce['thumbnail'], thumbnail_data)
 
         # Validate thumbnail content length
@@ -1168,8 +1099,8 @@ class VideoMessage(Message):
 
         # Download and decrypt video
         video_id = binascii.hexlify(video_id).decode('ascii')
-        response = yield from connection.download(video_id)
-        video_data = yield from response.read()
+        response = await connection.download(video_id)
+        video_data = await response.read()
         video = _sk_decrypt(key, cls.nonce['file'], video_data)
 
         # Validate video content length
@@ -1187,7 +1118,7 @@ class VideoMessage(Message):
 
 
 # TODO: Update docstring (arguments)
-@aio_run_proxy_decorator
+@aio_run_proxy
 class FileMessage(Message):
     """
     A file message including a thumbnail.
@@ -1310,15 +1241,14 @@ class FileMessage(Message):
                 # Read content
                 self._thumbnail_content = file.read()
 
-    @asyncio.coroutine
-    def pack(self, writer):
+    async def pack(self, writer):
         # Check capabilities of recipient
-        yield from self.check_capabilities(self.required_capabilities)
+        await self.check_capabilities(self.required_capabilities)
 
         # Encrypt and upload file by a newly generated symmetric key
         key, hex_key = Key.generate_secret_key()
         _, file_data = _sk_encrypt(key, self.file_content, nonce=self.nonce['file'])
-        file_id = yield from self._connection.upload(file_data)
+        file_id = await self._connection.upload(file_data)
 
         # Build JSON
         if self._file_path is not None:
@@ -1339,7 +1269,7 @@ class FileMessage(Message):
         if thumbnail_content is not None:
             _, thumbnail_data = _sk_encrypt(
                 key, thumbnail_content, nonce=self.nonce['thumbnail'])
-            thumbnail_id = yield from self._connection.upload(thumbnail_data)
+            thumbnail_id = await self._connection.upload(thumbnail_data)
             # Update JSON
             content['t'] = thumbnail_id
 
@@ -1353,8 +1283,7 @@ class FileMessage(Message):
         writer.writeexactly(content)
 
     @classmethod
-    @asyncio.coroutine
-    def unpack(cls, connection, parameters, key_pair, reader):
+    async def unpack(cls, connection, parameters, key_pair, reader):
         # Get payload
         try:
             content = bytes(reader.readexactly(len(reader))).decode('ascii')
@@ -1384,15 +1313,15 @@ class FileMessage(Message):
 
         # Download and decrypt thumbnail (if any)
         if thumbnail_id is not None:
-            response = yield from connection.download(thumbnail_id)
-            thumbnail_data = yield from response.read()
+            response = await connection.download(thumbnail_id)
+            thumbnail_data = await response.read()
             thumbnail_content = _sk_decrypt(key, cls.nonce['thumbnail'], thumbnail_data)
         else:
             thumbnail_content = None
 
         # Download and decrypt file
-        response = yield from connection.download(file_id)
-        file_data = yield from response.read()
+        response = await connection.download(file_id)
+        file_data = await response.read()
         file_content = _sk_decrypt(key, cls.nonce['file'], file_data)
 
         # Validate file content length
